import requests
import json
import dataclasses
from pyvise.graphy.base import get_logger
from pyvise.configuration import get_setting

logger = get_logger(__name__)


class GraphyException(Exception):
    pass


@dataclasses.dataclass
class GraphQLQuery:
    query: str = None
    # query_variables: dict = {}

    def generate_json(self, query_variables={}):
        return json.dumps({'query': self.query, 'variables': query_variables})

    def get_headers(self, secret=get_setting('HASURA_SECRET')):
        return {
            'content-type': 'application/json',
            'x-hasura-admin-secret': secret
        }


class GraphQLAPIClient:
    """
    Client transforms hasura responses to python code, throwing exceptions for errors
    """

    def __init__(self, endpoint: str = get_setting('HASURA_ENDPOINT'), secret: str = get_setting('HASURA_SECRET')):
        self.endpoint = endpoint
        self.secret = secret
        self.logger = get_logger(__name__)

    def execute_single_query(self, graphql_query: GraphQLQuery, query_variables: dict):
        response = requests.post(
            self.endpoint,
            headers=graphql_query.get_headers(self.secret),
            data=graphql_query.generate_json(query_variables=query_variables)
        )

        if(response.ok):
            response_message = response.json()

            if response_message.get('errors') is None:
                self.logger.info("GraphQL Response returned")
                return response_message.get('data')
            else:
                errors = response_message.get('errors')
                error_message = f'GraphQL returned an error! Server message: {errors}'
                self.logger.info(error_message, response_message)
                raise Exception(error_message, response_message)
        else:
            error_message = f"HTTP Exception {response.status_code}"
            self.logger.error(error_message)
            raise GraphyException(error_message, response)

    def execute_query(self, graphql_query: GraphQLQuery, operation_name, query_variables: dict):
        raise NotImplementedError("This hasn't been implemented yet")
        # TODO: Add operation name to JSON generation multi_query = query_variables

        response = requests.post(
            self.endpoint,
            headers=graphql_query.get_headers(self.secret),
            data=graphql_query.generate_json(query_variables=query_variables)
        )

        if(response.status_code != 200):
            response_message = response.json()

            if response_message.get('errors') is None:
                self.logger.info("GraphQL Response returned")
                return response_message.get('data')
            else:
                errors = response_message.get('errors')
                error_message = f'GraphQL returned an error! Server message: {errors}'
                self.logger.info(error_message, response_message)
                raise Exception(error_message, response_message)
        else:
            error_message = f"HTTP Exception {response.status_code}"
            self.logger.error(error_message, response)
            raise Exception(error_message, response)



import datetime
import logging
import sys
from fastapi_jwt_auth import AuthJWT
from .configuration import get_setting, get_AuthJWT_instance

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
def get_logger(name):
    return logging.getLogger(name)


def create_jwt_token(Authorize: AuthJWT, did_token):
    access_token = Authorize.create_access_token(subject=did_token)
    return access_token


def create_jwt_token_by_email(Authorize: AuthJWT, email_address, valid_for=365):
    in_1_year = datetime.timedelta(days=valid_for)
    access_token = Authorize.create_access_token(subject=email_address, expires_time=in_1_year)
    return {
        'access_token': access_token,
        'expiration': in_1_year
    }

def decode_jwt(token: str) -> dict:
    instance = get_AuthJWT_instance()

    result = instance.get_raw_jwt(encoded_token=token)
    """
    {'sub': 'dennis@payonk.com', 
    'iat': 1611082366, 
    'nbf': 1611082366, 
    'jti': 'd602bcf3-a367-42be-8d3f-b2ac6601343a', 
    'exp': 1611083266, 
    'type': 'access', 
    'fresh': False}
    """
    print("TODO: check expiration")
    return result['sub']


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("VERSION", "r", encoding="utf-8") as fh:
    version = fh.read()

setuptools.setup(
    name="pyvise", # Replace with your own username
    version=version,
    author="Dennis Payonk",
    author_email="dennis@payonk.com",
    description="A small authentication package using FastAPI and magic",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dpayonk/pyvise",
    packages=setuptools.find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        'aiofiles>=0.6.0',
        'dataclasses-json>=0.5.2',
        'fastapi>=0.63.0',
        'fastapi-jwt-auth>=0.5.0',
        'pydantic>=1.7.3',
        'magic-admin>=0.0.4',
        'requests>=2.22.0'
    ]
)


import os
import logging
import sys
from pydantic import BaseModel
from fastapi_jwt_auth import AuthJWT

logger = logging.getLogger(__name__)
if (os.environ.get('ENV') == 'development'):
    # Using stdout logging in development mode
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)


APP_CONFIGS = {}  # SINGLETON Variable to hold configurations
DEFAULTS__MAGIC_SECRET_KEY = os.environ.get('MAGIC_SECRET_KEY') # from magic
DEFAULTS__JWT_SECRET = os.environ.get('JWT_SECRET') # token to authorize calls

# Used by GraphQL backend to persist data
DEFAULTS__HASURA_ENDPOINT = os.environ.get('HASURA_ENDPOINT')
DEFAULTS__HASURA_SECRET = os.environ.get('HASURA_SECRET')


def get_setting(key):
    if APP_CONFIGS.get(key) is not None:
        return APP_CONFIGS.get(key)
    # This should throw an exception
    if key == 'HASURA_ENDPOINT':        
        return DEFAULTS__HASURA_ENDPOINT
    elif key == 'HASURA_SECRET':
        return DEFAULTS__HASURA_SECRET
    elif key == 'MAGIC_SECRET_KEY':
        return DEFAULTS__MAGIC_SECRET_KEY
    elif key == 'JWT_SECRET':
        return DEFAULTS__JWT_SECRET

    return None


JWT_SINGLETON = None

def configure(**kwargs):
    """
    Iterate over configurations to override APP_CONFIGS and return an AuthJWT object
    Should set MAGIC_SECRET_KEY, JWT_SECRET, HASURA_SECRET, HASURA_ENDPOINT
    By default, it will try and pull environment variables
    """
    global JWT_SINGLETON
    global APP_CONFIGS

    for k, v in kwargs.items():
        APP_CONFIGS[k] = v

    if(len(APP_CONFIGS.items()) < 4):
        logger.warning('All configurations have not been set')

    logger.info("Configuring AuthJWT")
    @AuthJWT.load_config
    class Config(BaseModel):
        authjwt_secret_key: str = get_setting('JWT_SECRET')

    JWT_SINGLETON = AuthJWT()
    return JWT_SINGLETON


def get_AuthJWT_instance():
    global JWT_SINGLETON
    if JWT_SINGLETON is None:
        logger.warning('AuthJWT has not been initialied, using default')
        JWT_SINGLETON = configure()

    return JWT_SINGLETON



from datetime import datetime
from fastapi import HTTPException, Depends
from fastapi import APIRouter
from pydantic import BaseModel
from pyvise.base import AuthJWT, create_jwt_token_by_email, get_logger
from pyvise.backend.magic_backend import find_magic_user
from pyvise.backend.account_backend import create_account_profile

profile_router = APIRouter()
logger = get_logger(__name__)


class CreateProfileRequestSchema(BaseModel):
    email_address: str
    did_token: str = None

    def generate_response(self, email_address, did_token, magic_user, jwt_token):
        return {
            'data': {
                'email_address': email_address,
                'jwt_token': jwt_token,
                'did_token': did_token,
                'user': magic_user,
            },
            'errors': ""
        }


@profile_router.post('/create',  tags=["account_profile", "public"])
def create_url(req_schema: CreateProfileRequestSchema, Authorize: AuthJWT = Depends()):
    """
    Validates user against magic (using didToken), creates backend record
    and returns a JWT token
    
    [Reference Design](https://docs.magic.link/admin-sdk/python/examples/user-signup)
    - Step 1 of signup, called after onRedirectLogin
    # 
    """
    try:
        valid_user = find_magic_user(
            req_schema.email_address, req_schema.did_token)

        if(valid_user is not None):
            jwt_token_dict = create_jwt_token_by_email(
                Authorize,
                req_schema.email_address, 365)

            create_account_profile(
                valid_user.key, req_schema.email_address,
                jwt_token_dict['access_token'],
                jwt_token_dict['expiration'])

            return req_schema.generate_response(
                req_schema.email_address,
                req_schema.did_token,
                valid_user,
                jwt_token_dict['access_token']
            )
        else:
            return {'errors': 'Could not validate user or create profile'}
    except Exception as magicDIDTokenError:
        # raise BadRequest('DID Token is invalid: {}'.format(magicDIDTokenError))
        raise HTTPException(
            status_code=401, detail=f"Invalid DID Token likely. Error: {magicDIDTokenError}")
